/*
   vda2fbd: convert from VDA/FS to fbd format
   Copyright 1995 Johannes Barner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published  by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <stdlib.h>
#include <math.h>

/* from the book "Numerical Recipes" */

float LnGamma(float xx)
{
   int j;
   double x,y,tmp,ser;
   double cof[6] = {
      76.18009172947146,    -86.50532032941677,
      24.01409824083091,    -1.231739572450155,
      0.1208650973866179e-2,-0.5395239384953e-5
   };

   y = x = xx;
   tmp = x + 5.5 - (x + 0.5) * log(x + 5.5);
   ser = 1.000000000190015;
   for (j=0; j<=5; j++)
      ser += (cof[j] / ++y);
   return(log(2.5066282746310005 * ser / x) - tmp);
}

float factln(int n)
{
    static float a[101];

    if (n <= 1) return 0.0;
    if (n <= 100) return a[n] ? a[n] : (a[n]=LnGamma(n+1.0));
    else return LnGamma(n+1.0);
}


float binom_coef(int n, int k)
{
    return floor(0.5+exp(factln(n)-factln(k)-factln(n-k)));
}


void	pol2bez_curve(	double	*a,
			int	m,
			double	**LT_i,
			double	*c )
{
    int	i, j;
    double	sum;

    for( i = 0; i < m; i++ )
        for( j = i; j < m; j++ )
            LT_i[j][i] = (double)( binom_coef( j, i )/binom_coef( m-1, i ) );
    
    for( i = 0; i < m; i++ ) {
        for( j = 0, sum = 0.; j <= i; j++ )
            sum = sum + LT_i[i][j]*a[j];
        c[i] = sum;
    }
    
    for( i = 0; i < m; i++ )
        a[i] = c[i];
}



void	pol2bez_surf(	double	**A,
			int	m,
			int	n,
			double	**MT_i,
			double	**N_i,
			double	**b )
{
    int	i, j, k;
    double	coef;

    for( j = 0; j < n; j++ )
        for( i = 0, coef = A[0][j]; i < m; i++ )
            b[i][j] = coef;
    for( i = 1; i < m; i++ )
        for( j = 1; j <= i; j++ )
            for( k = 0, coef = binom_coef(i,j)/binom_coef(m-1,j); k < n; k++ )
                b[i][k] += coef*A[j][k];
    
    for( i = 0; i < m; i++ )
        for( j = 0, coef = b[i][0]; j < n; j++ )
            A[i][j] = coef;
    for( j = 1; j < n; j++ )
        for( i = 1; i <= j; i++ )
            for( k = 0, coef = binom_coef(j,i)/binom_coef(n-1,i); k < m; k++ )
                A[k][j] += coef*b[k][i];
}
