/*
   vda2fbd: convert from VDA/FS to fbd format
   Copyright 1995 Johannes Barner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published  by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	"vda.h"

extern	int   max_u_order, max_v_order;
extern	int	n_points;
extern	int	n_curves;
extern	int	n_surfaces;
extern	int	n_patches;

extern	VDA_SET		enqueue_vdaset();
extern	VDA_HEADER	enqueue_header();
extern	VDA_POINT	enqueue_point();
extern	VDA_PSET	enqueue_pset();
extern	VDA_MDI		enqueue_mdi();
extern	VDA_CURVE	enqueue_curve();
extern	POLY_1		enqueue_poly1();
extern	VDA_CIRCLE	enqueue_circle();
extern	VDA_SURFACE	enqueue_surface();
extern	POLY_2		enqueue_poly2();
extern	VDA_FACE	enqueue_face();
extern	VDA_CONS	enqueue_cons();
extern	VDA_TOP		enqueue_top();
extern	VDA_GROUP	enqueue_group();
extern	VDA_TMAT	enqueue_tmat();
extern	VDA_TLIST	enqueue_tlist();
extern	CONS_CURVE	enqueue_cons_curve();
extern	FACE_CONS	enqueue_face_cons();
extern	FACE_LOOP	enqueue_face_loop();
extern	TOP_CONS	enqueue_top_cons();
extern	TOP_FACE	enqueue_top_face();

extern	VDA_SURFACE	find_vda_surface();
extern	VDA_CURVE	find_vda_curve();
extern	VDA_CONS	find_vda_cons();

VDA	init_vda(	char	*name	)
{
	VDA	model;

	model = (VDA)malloc( sizeof( struct vda ) );
	model->name = (char*)malloc( 1+strlen( name ) );
	strcpy( model->name, name );
	model->set = (VDA_SET)malloc( sizeof( struct vda_set ) );
	model->set->next = NULL;
	return model;
}

VDA_SET	init_vdaset(	VDA	model,
			char	*name )
{
	VDA_SET	set;

	set = (VDA_SET)malloc( sizeof( struct vda_set ) );
	set->name = (char*)malloc( 1+strlen( name ) );
	strcpy( set->name, name );
	set->next = NULL;

	set->header = (VDA_HEADER)malloc( sizeof( struct vda_header ) );
	set->header->key   = NULL;
	set->header->value = NULL;
	set->header->next  = NULL;

	set->point = (VDA_POINT)malloc( sizeof( struct vda_point ) );
	set->point->name = NULL;
	set->point->c    = NULL;
	set->point->next = NULL;

	set->pset = (VDA_PSET)malloc( sizeof( struct vda_pset ) );
	set->pset->name = NULL;
	set->pset->n    = 0;
	set->pset->c    = NULL;
	set->pset->next = NULL;

	set->mdi = (VDA_MDI)malloc( sizeof( struct vda_mdi ) );
	set->mdi->name = NULL;
	set->mdi->n    = 0;
	set->mdi->c    = NULL;
	set->mdi->next = NULL;

	set->circle = (VDA_CIRCLE)malloc( sizeof( struct vda_circle ) );
	set->circle->name   = NULL;
	set->circle->params = NULL;
	set->circle->next   = NULL;

	set->curve = (VDA_CURVE)malloc( sizeof( struct vda_curve ) );
	set->curve->name = NULL;
	set->curve->ns   = 0;
	set->curve->a    = NULL;
	set->curve->poly = NULL;
	set->curve->next = NULL;

	set->surface = (VDA_SURFACE)malloc( sizeof( struct vda_surface ) );
	set->surface->name = NULL;
	set->surface->ns   = 0;
	set->surface->nt   = 0;
	set->surface->a    = NULL;
	set->surface->poly = NULL;
	set->surface->next = NULL;

	set->cons  = (VDA_CONS)malloc( sizeof( struct vda_cons ) );
	set->cons->name           = NULL;
	set->cons->surf           = NULL;
	set->cons->curve          = NULL;
	set->cons->global_params  = NULL;
	set->cons->n_segments     = 0;
	set->cons->segment_params = NULL;
	set->cons->cons_curve     = NULL;
	set->cons->next           = NULL;

	set->face  = (VDA_FACE)malloc( sizeof( struct vda_face ) );
	set->face->identifier = NULL;
	set->face->name       = NULL;
	set->face->surf       = NULL;
	set->face->n_loops    = 0;
	set->face->loop       = NULL;
	set->face->next       = NULL;

	set->top   = (VDA_TOP)malloc( sizeof( struct vda_top ) );
	set->top->n_face_pairs = 0;
	set->top->face         = NULL;
	set->top->next         = NULL;

	set->group = (VDA_GROUP)malloc( sizeof( struct vda_group ) );
	set->group->name        = NULL;
	set->group->n_objects   = 0;
	set->group->object_list = NULL;
	set->group->next        = NULL;

	set->tmat  = (VDA_TMAT)malloc( sizeof( struct vda_tmat ) );
	set->tmat->name   = NULL;
	set->tmat->params = NULL;
	set->tmat->next   = NULL;

	set->tlist = (VDA_TLIST)malloc( sizeof( struct vda_tlist ) );
	set->tlist->name        = NULL;
	set->tlist->n_objects   = 0;
	set->tlist->object_list = NULL;
	set->tlist->next        = NULL;

	enqueue_vdaset( model, set );

	return set;
}

void	init_header(	VDA_SET	set,
			char	*key,
			char	*value )
{
	VDA_HEADER	header;

	header = (VDA_HEADER)malloc( sizeof( struct vda_header ) );
	header->key = (char*)malloc( 1+strlen( key ) );
	strcpy( header->key, key );
	header->value = (char*)malloc( 1+strlen( value ) );
	strcpy( header->value, value );
	header->next = NULL;
	enqueue_header( set, header );
}

double 	*init_point(	VDA_SET	set,
			char	*name )
{
	VDA_POINT	point;
	extern	int	n_points;

	n_points++;
	point = (VDA_POINT)malloc( sizeof( struct vda_point ) );
	point->name = (char*)malloc( 1+strlen( name ) );
	strcpy( point->name, name );
	point->c = (double*)malloc( 3*sizeof( double ) );
	point->next = NULL;
	enqueue_point( set, point );
	return point->c;
}

double 	*init_pset(	VDA_SET	set,
			char	*name,
			int	n )
{
	VDA_PSET	pset;
	extern	int	n_points;

	n_points += n;
	pset = (VDA_PSET)malloc( sizeof( struct vda_pset ) );
	pset->name = (char*)malloc( 1+strlen( name ) );
	strcpy( pset->name, name );
	pset->n = n;
	pset->c = (double*)malloc( 3*n*sizeof( double ) );
	pset->next = NULL;
	enqueue_pset( set, pset );
	return pset->c;
}

double 	*init_circle(	VDA_SET	set,
			char	*name )
{
	VDA_CIRCLE	circle;

	circle = (VDA_CIRCLE)malloc( sizeof( struct vda_circle ) );
	circle->name = (char*)malloc( 1+strlen( name ) );
	strcpy( circle->name, name );
	circle->params = (double*)malloc( 12*sizeof( double ) );
	circle->next = NULL;
	enqueue_circle( set, circle );
	return circle->params;
}

double 	*init_mdi(	VDA_SET	set,
			char	*name,
			int	n )
{
	VDA_MDI	mdi;

	mdi = (VDA_MDI)malloc( sizeof( struct vda_mdi ) );
	mdi->name = (char*)malloc( 1+strlen( name ) );
	strcpy( mdi->name, name );
	mdi->n = n;
	mdi->c = (double*)malloc( 6*n*sizeof( double ) );
	mdi->next = NULL;
	enqueue_mdi( set, mdi );
	return mdi->c;
}

double 	*init_curve(	VDA_SET	set,
			char	*name,
			int	ns )
{
	VDA_CURVE	curve;
	extern	int	n_curves;

	n_curves++;
	curve = (VDA_CURVE)malloc( sizeof( struct vda_curve ) );
	curve->name = (char*)malloc( 1+strlen( name ) );
	strcpy( curve->name, name );
	curve->next = NULL;
	curve->ns = ns;
	curve->a = (double*)malloc( ( ns + 1 )*sizeof( double ) );
	curve->poly = (POLY_1)malloc( sizeof( struct vda_poly_1d ) );
	curve->poly->nu   = 0;
	curve->poly->a    = NULL;
	curve->poly->next = NULL;
	enqueue_curve( set, curve );
	return curve->a;
}

static long lmaxarg1,lmaxarg2;
#define LMAX(a,b) (lmaxarg1=(a),lmaxarg2=(b),(lmaxarg1) > (lmaxarg2) ?\
        (lmaxarg1) : (lmaxarg2))

double 	*init_curve_poly(	VDA_SET	set,
				int	nu )
{
	POLY_1	poly1;
	double	*a;
	extern	int	max_u_order;

        max_u_order = LMAX( max_u_order, nu );

	poly1 = (POLY_1)malloc( sizeof( struct vda_poly_1d ) );
	poly1->nu = nu;
	a = (double*)malloc( 3*nu*sizeof( double ) );
	poly1->a = a;
	poly1->next = NULL;
	enqueue_poly1( set->curve, poly1 );
	return a;
}

double 	*init_surface(	VDA_SET	set,
			char	*name,
			int	ns,
			int	nt )
{
	VDA_SURFACE	surface;
	extern	int	n_surfaces, n_patches;

	n_surfaces++;
	n_patches += ns*nt;
	surface = (VDA_SURFACE)malloc( sizeof( struct vda_surface ) );
	surface->name = (char*)malloc( 1+strlen( name ) );
	strcpy( surface->name, name );
	surface->ns = ns;
	surface->nt = nt;
	surface->next = NULL;
	surface->a = (double*)malloc( ( ns + 1 )*( nt + 1 )*sizeof( double ) );
	surface->poly = (POLY_2)malloc( sizeof( struct vda_poly_2d ) );
	surface->poly->nu   = 0;
	surface->poly->nv   = 0;
	surface->poly->a    = NULL;
	surface->poly->next = NULL;
	enqueue_surface( set, surface );
	return surface->a;
}

double 	*init_surf_poly(	VDA_SET	set,
				int	nu,
				int	nv )
{
	POLY_2	poly2;
	double	*a;
	extern	int	max_u_order, max_v_order;

        max_u_order = LMAX( max_u_order, nu );
        max_v_order = LMAX( max_v_order, nv );

	poly2 = (POLY_2)malloc( sizeof( struct vda_poly_2d ) );
	poly2->nu = nu;
	poly2->nv = nv;
	a = (double*)malloc( 3*nu*nv*sizeof( double ) );
	poly2->a = a;
	poly2->next = NULL;
	enqueue_poly2( set->surface, poly2 );
	return a;
}



double 	*init_cons_param(	VDA_SET	set,
				int	n_segments )
{
	VDA_CONS	cons = set->cons;
	CONS_CURVE	cons_curve;
	double		*segment_params;

	while( cons->next != NULL )
		cons = cons->next;
	cons_curve = cons->cons_curve;
	while( cons_curve->next != NULL )
		cons_curve = cons_curve->next;
	cons->n_segments = n_segments;
	segment_params = (double*)malloc( (n_segments+1)*sizeof( double ) );
	cons->segment_params = segment_params;
	return segment_params;
}

double 	*init_cons_curve(	VDA_SET	set,
				int	order )
{
	CONS_CURVE	cons_curve;
	double		*coeff;

	cons_curve = (CONS_CURVE)malloc( sizeof( struct vda_cons_curve ) );
	cons_curve->order = order;
	coeff = (double*)malloc( 2*order*sizeof( double ) );
	cons_curve->coeff = coeff;
	cons_curve->next = NULL;
	enqueue_cons_curve( set->cons, cons_curve );
	return coeff;
}

double 	*init_cons(	VDA_SET	set,
			char	*name,
			char	*surf_name,
			char	*curve_name )
{
	int		i = 0;
	VDA_CONS	cons;
	double		*params;

	cons = (VDA_CONS)malloc( sizeof( struct vda_cons ) );
	cons->name = (char*)malloc( 1+strlen( name ) );
	strcpy( cons->name, name );
	cons->surf = find_vda_surface( set->surface, surf_name );
	cons->curve = find_vda_curve( set->curve, curve_name );
	params = (double*)malloc( 2*sizeof( double ) );
	cons->global_params = params;
	cons->n_segments = 0;
	cons->segment_params = NULL;
	cons->cons_curve = (CONS_CURVE)malloc( sizeof(struct vda_cons_curve) );
	cons->cons_curve->order = 0;
	cons->cons_curve->coeff = NULL;
	cons->cons_curve->next = NULL;
	cons->next = NULL;
	i = strlen(surf_name);
	enqueue_cons( set, cons );
	return params;
}



double 	*init_face_cons(	VDA_SET	set,
				char	*cons_name )
{
	FACE_CONS	face_cons;
	double		*params;

	face_cons = (FACE_CONS)malloc( sizeof( struct vda_face_cons ) );
	face_cons->cons = find_vda_cons( set->cons, cons_name );
	params = (double*)malloc( 2*sizeof( double ) );
	face_cons->params = params;
	face_cons->next = NULL;
	enqueue_face_cons( set->face, face_cons );
	return params;
}

void 	init_face_loop(	VDA_SET	set,
			int	n_cons )
{
	FACE_LOOP	loop;

	loop = (FACE_LOOP)malloc( sizeof( struct vda_face_loop ) );
	loop->n_cons = n_cons;
	loop->face_cons = (FACE_CONS)malloc( sizeof( struct vda_face_cons ) );
	loop->face_cons->params = NULL;
	loop->face_cons->next = NULL;
	loop->next = NULL;
	enqueue_face_loop( set->face, loop );
}

void 	init_face(	VDA_SET	set,
			char	*identifier,
			char	*name,
			char	*surf_name,
			int	n_loops )
{
	VDA_FACE	face;

	face = (VDA_FACE)malloc( sizeof( struct vda_face ) );
	face->identifier = (char*)malloc( 1+strlen( identifier ) );
	strcpy( face->identifier, identifier );
	face->name = (char*)malloc( 1+strlen( name ) );
	strcpy( face->name, name );
	face->surf = find_vda_surface( set->surface, surf_name );
	face->n_loops = n_loops;
	face->loop = (FACE_LOOP)malloc( sizeof( struct vda_face_loop ) );
	face->loop->n_cons = 0;
	face->loop->face_cons = NULL;
	face->loop->next = NULL;
	face->next = NULL;
	enqueue_face( set, face );
}



double 	*init_top_cons(	VDA_SET	set,
			char	*name )
{
	TOP_CONS	cons;
	double		*global_params;

	cons = (TOP_CONS)malloc( sizeof( struct vda_top_cons ) );
	cons->name = (char*)malloc( 1+strlen( name ) );
	strcpy( cons->name, name );
	global_params = (double*)malloc( 2*sizeof( double ) );
	cons->global_params = global_params;
	cons->next = NULL;
	enqueue_top_cons( set->top, cons );
	return global_params;
}

void 	init_top_face(	VDA_SET	set,
			char	*name,
			int	n_curves )
{
	TOP_FACE	face;

	face = (TOP_FACE)malloc( sizeof( struct vda_top_face ) );
	face->n_curves = n_curves;
	face->cons = (TOP_CONS)malloc( sizeof( struct vda_top_cons ) );
	face->cons->global_params = NULL;
	face->cons->next = NULL;
	face->cont_flag = 0;
	face->cons->next = NULL;
	face->next = NULL;
	enqueue_top_face( set->top, face );
}

void 	init_top(	VDA_SET	set,
			int	n_face_pairs )
{
	VDA_TOP		top;

	top = (VDA_TOP)malloc( sizeof( struct vda_top ) );
	top->n_face_pairs = 0;
	top->face = (TOP_FACE)malloc( sizeof( struct vda_top_face ) );
	top->face->n_curves = 0;
	top->face->cons = NULL;
	top->face->cont_flag = 0;
	top->face->next = NULL;
	top->next = NULL;
	enqueue_top( set, top );
}



CHARPTR	*init_group(	VDA_SET	set,
			char	*name,
			int	n_objects )
{
	VDA_GROUP	group;
	CHARPTR		*object_list;

	group = (VDA_GROUP)malloc( sizeof( struct vda_group ) );
	group->name = (char*)malloc( 1+strlen( name ) );
	strcpy( group->name, name );
	group->n_objects = n_objects;
	object_list = (CHARPTR*)malloc( n_objects*sizeof( CHARPTR ) );
	group->object_list = object_list;
	group->next = NULL;
	enqueue_group( set, group );
	return object_list;
}



double 	*init_tmat(	VDA_SET	set,
			char	*name )
{
	VDA_TMAT	tmat;

	tmat = (VDA_TMAT)malloc( sizeof( struct vda_tmat ) );
	tmat->name = (char*)malloc( 1+strlen( name ) );
	strcpy( tmat->name, name );
	tmat->params = (double*)malloc( 12*sizeof( double ) );
	tmat->next = NULL;
	enqueue_tmat( set, tmat );
	return tmat->params;
}



CHARPTR *init_tlist(	VDA_SET	set,
			char	*name,
			int	n_objects )
{
	VDA_TLIST	tlist;
	CHARPTR		*object_list;

	tlist = (VDA_TLIST)malloc( sizeof( struct vda_tlist ) );
	tlist->name = (char*)malloc( 1+strlen( name ) );
	strcpy( tlist->name, name );
	tlist->n_objects = n_objects;
	object_list = (CHARPTR*)malloc( n_objects*sizeof( CHARPTR ) );
	tlist->object_list = object_list;
	tlist->next = NULL;
	enqueue_tlist( set, tlist );
	return	object_list;
}
