/*
   vda2fbd: convert from VDA/FS to fbd format
   Copyright 1995 Johannes Barner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published  by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */


#define FALSE 0
#define TRUE  1

typedef	char	*CHARPTR;

typedef	struct	vda_show	*VDA_SHOW;
typedef	struct	vda		*VDA;
typedef	struct	vda_set		*VDA_SET;

typedef	struct	vda_header	*VDA_HEADER;
typedef	struct	vda_point	*VDA_POINT;
typedef	struct	vda_pset	*VDA_PSET;
typedef	struct	vda_mdi		*VDA_MDI;
typedef	struct	vda_circle	*VDA_CIRCLE;
typedef	struct	vda_curve	*VDA_CURVE;
typedef	struct	vda_surface	*VDA_SURFACE;
typedef	struct	vda_cons	*VDA_CONS;
typedef	struct	vda_face	*VDA_FACE;
typedef	struct	vda_top		*VDA_TOP;
typedef	struct	vda_group	*VDA_GROUP;
typedef	struct	vda_tmat	*VDA_TMAT;
typedef	struct	vda_tlist	*VDA_TLIST;

typedef	struct	vda_poly_1d	*POLY_1;
typedef	struct	vda_poly_2d	*POLY_2;
typedef	struct	vda_cons_curve	*CONS_CURVE;
typedef	struct	vda_face_cons	*FACE_CONS;
typedef	struct	vda_face_loop	*FACE_LOOP;
typedef	struct	vda_top_cons	*TOP_CONS;
typedef	struct	vda_top_face	*TOP_FACE;

/*********************************************************************/

struct vda_show
{
    int header;
    int points;
    int curves;
    int surfaces;
    int patches;
    int faces;
    int layers;
    char* symbol;
};

struct	vda_header
{
	char		*key;
	char		*value;
	VDA_HEADER	next;
};

struct	vda_point
{
	char		*name;
	double		*c;
	VDA_POINT	next;
};

struct	vda_pset
{
	char		*name;
	int		n;
	double		*c;
	VDA_PSET	next;
};

struct	vda_mdi
{
	char		*name;
	int		n;
	double		*c;
	VDA_MDI		next;
};

struct	vda_circle
{
	char		*name;
	double		*params;
	VDA_CIRCLE	next;
};

struct	vda_group
{
	char		*name;
	int		n_objects;
	CHARPTR		*object_list;
	VDA_GROUP	next;
};

struct	vda_tmat
{
	char		*name;
	double		*params;
	VDA_TMAT	next;
};

struct	vda_tlist
{
	char		*name;
	int		n_objects;
	CHARPTR		*object_list;
	VDA_TLIST	next;
};

/*********************************************************************/

struct	vda_poly_1d
{
	int		nu;
	double		*a;
	POLY_1		next;
};

struct	vda_curve
{
	char		*name;
	int		ns;
	double		*a;
	POLY_1		poly;
	VDA_CURVE	next;
};

/*********************************************************************/

struct	vda_poly_2d
{
	int		nu, nv;
	double		*a;
	POLY_2		next;
};

struct	vda_surface
{
	char		*name;
	int		ns, nt;
	double		*a;
	POLY_2		poly;
	VDA_SURFACE	next;
};

/*********************************************************************/

struct	vda_cons_curve
{
	int		order;
	double		*coeff;
	CONS_CURVE	next;
};

struct	vda_cons
{
	char		*name;
	VDA_SURFACE	surf;
	VDA_CURVE	curve;
	double		*global_params;
	int		n_segments;
	double		*segment_params;
	CONS_CURVE	cons_curve;
	VDA_CONS	next;
};

/*********************************************************************/

struct	vda_face_cons
{
	VDA_CONS	cons;
	double		*params;
	FACE_CONS	next;
};

struct	vda_face_loop
{
	int		n_cons;
	FACE_CONS	face_cons;
	FACE_LOOP	next;
};

struct	vda_face
{
	char		*identifier;
	char		*name;
	VDA_SURFACE	surf;
	int		n_loops;
	FACE_LOOP	loop;
	VDA_FACE	next;
};

/*********************************************************************/

struct	vda_top_cons
{
	char		*name;
	double		*global_params;
	TOP_CONS	next;
};

struct	vda_top_face
{
	char		*name;
	int		n_curves;
	TOP_CONS	cons;
	int		cont_flag;
	TOP_FACE	next;
};

struct	vda_top
{
	int		n_face_pairs;
	TOP_FACE	face;
	VDA_TOP		next;
};

/*********************************************************************/

struct	vda_set
{
	char		*name;
	VDA_HEADER	header;
	VDA_POINT	point;
	VDA_PSET	pset;
	VDA_MDI		mdi;
	VDA_CURVE	curve;
	VDA_CIRCLE	circle;
	VDA_SURFACE	surface;
	VDA_CONS	cons;
	VDA_FACE	face;
	VDA_TOP		top;
	VDA_GROUP	group;
	VDA_TMAT	tmat;
	VDA_TLIST	tlist;
	VDA_SET		next;
};

struct	vda
{
	char		*name;
	VDA_SET		set;
};

/*********************************************************************/
