/*
   vda2fbd: convert from VDA/FS to fbd format
   Copyright 1995 Johannes Barner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published  by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

%{
#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	"vda.h"
#include	"vda_util.h"

CHARPTR		*s_ptr;
double		*r_ptr;
VDA		model;
VDA_SET		set;
int		line_number;
char		yitem[80], ysurf[80], ycurve[80];
%}

%union
{	char	*s, *t, *v;
	int	i;
	double	r;
}

%start	list
%token	HEADER BEGINSET GROUP TMAT TLIST ENDSET END
%token	POINT PSET MDI CIRCLE CURVE SURF CONS FACE TOP
%token	<t>	ITEM
%token	<s>	NAME
%token	<v>	IDENTIFIER
%token	<i>	INT
%token	<r>	REAL

%%

list:
	| list expr;

expr:	header
	| seto
	| point  real_list
	| pset   real_list
	| mdi    real_list
	| circle real_list
	| curve  curve_list
	| surf   surf_list
	| cons   cons_list
	| face   face_list
	| cface  face_list
	| top    top_list
	| group  name_list
	| tmat   real_list
	| tlist  name_list
	| setc
	| end;

header:	ITEM HEADER		{ init_vdaset( model, $1 ); set = model->set->next; };
seto:	ITEM BEGINSET		{};
point:	ITEM POINT 		{ r_ptr = init_point(   set, $1         ); };
pset:	ITEM PSET  INT		{ r_ptr = init_pset(    set, $1, $3     ); };
mdi:	ITEM MDI   INT		{ r_ptr = init_mdi(     set, $1, $3     ); };
circle:	ITEM CIRCLE		{ r_ptr = init_circle(  set, $1         ); };
curve:	ITEM CURVE INT		{ r_ptr = init_curve(   set, $1, $3     ); };
surf:	ITEM SURF  INT  INT 	{ r_ptr = init_surface( set, $1, $3, $4 ); };
cons:	ITEM CONS  NAME         { strcpy( yitem, $1 );  strcpy( ysurf, $3 ); };
face:	ITEM FACE  NAME INT 	{         init_face(    set, "\0", $1, $3, $4 ); };
cface:	IDENTIFIER ITEM FACE  NAME INT 	{ init_face(    set, $1, $2, $4, $5 ); };
top:	ITEM TOP   INT		{         init_top(     set, $1, $3     ); };
tmat:	ITEM TMAT	  	{ r_ptr = init_tmat(    set, $1         ); };
tlist:	ITEM TLIST NAME INT	{ s_ptr = init_tlist(   set, $1, $3, $4 ); };
group:	ITEM GROUP INT		{ s_ptr = init_group(   set, $1, $3     ); };
setc:	ITEM ENDSET		{};
end:	ITEM END		{};

curve_list:	real_list segment;
segment:	order real_list | segment order real_list;
order:		INT	{ r_ptr = init_curve_poly( set, $1 ); };

surf_list:	real_list patch;
patch:		orders real_list | patch orders real_list;
orders:		INT INT	{ r_ptr = init_surf_poly(  set, $1, $2 ); };

cons_list:	curve_name real real n_curves real_list curve_ons;
curve_ons:	c_order real_list | curve_ons c_order real_list;
curve_name:	NAME	{	strcpy( ycurve, $1 );
				r_ptr = init_cons( set, yitem, ysurf, ycurve ); };
n_curves:	INT 	{ r_ptr = init_cons_param( set, $1 ); };
c_order:	INT 	{ r_ptr = init_cons_curve( set, $1 ); };

face_list:	loop_list | face_list loop_list;
loop_list:	n_face_cons face_cons_list;
face_cons_list:	cons_name real_list | face_cons_list cons_name real_list;
n_face_cons:	INT 	{         init_face_loop( set, $1 ); };
cons_name:	NAME	{ r_ptr = init_face_cons(  set, $1 ); };

top_list:	pair_list | top_list pair_list;
pair_list:	top_face top_cons_list top_face top_cons_list icont;
top_cons_list:	top_cons_name real real | top_cons_list top_cons_name real real;
top_face:	NAME INT {         init_top_face( set, $1, $2 ); };
top_cons_name:	NAME	{ r_ptr = init_top_cons( set, $1 ); };
icont:		INT	{};

name_list:	name | name_list name;
name:		NAME	{	*s_ptr = (char*)malloc( 1 + strlen( $1 ) );
				strcpy( *s_ptr++, $1 );
			};

real_list:	real | real_list real;
real:		REAL	{ *r_ptr++ = $1; };

%%

int	yyerror( char *s )
{	fprintf( stderr, "yacc: %s  Line%8d\n", s, line_number+1 );
	return( 1 );
}

int	yywrap()
{	return( 1 ) ;
}




