/*
   vda2fbd: convert from VDA/FS to fbd format
   Copyright 1995 Johannes Barner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published  by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<math.h>

#include	"vda.h"
#include	"fbd.h"

double *double_vector(int n0, int n1)
{
	double *v;

	v=(double *)malloc((n1 - n0 + 2)*sizeof(double));
	return v - n0 + 1;
}

double **double_matrix(int m0, int m1, int n0, int n1)
{
	int i, nrow=m1 - m0 + 1, ncol = n1 - n0 + 1;
	double **m;

	m=(double **)malloc((nrow+1)*sizeof(double*));
	m += 1;
	m -= m0;

	m[m0]=(double *) malloc((nrow*ncol+1)*sizeof(double));
	m[m0] += 1;
	m[m0] -= n0;

	for(i = m0 + 1; i <= m1; i++) m[i] = m[i-1]+ncol;

	return m;
}

void free_double_vector(double *v, int n0, int n1)
{
	free(v+n0-1);
}

void free_double_matrix(double **m, int m0, int m1, int n0, int n1)
{
	free(m[m0] + n0 - 1);
	free(m + m0 - 1);
}

#define IMIN(arg1,arg2) (arg1 < arg2 ? arg1 : arg2)
#define LMAX(arg1,arg2) (arg1 > arg2 ? arg1 : arg2)

extern void	pol2bez_curve(	double	*a,
				int	m,
				double	**LT_i,
				double	*c );
extern void	pol2bez_surf(	double	**A,
				int	m,
				int	n,
				double	**MT_i,
				double	**N_i,
				double	**b );

extern int max_u_order, max_v_order;

char	unit[5] = "mm";
double	scale = 1.0;
double	gtol  = 0.001;

const   int	fam_n_chars = 36;
char	*fam_chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

char	fam_number[4], tmp_set[5];

char    first_symbol_name[5];
char    fam_symbol_name[5];
char    tmp_symbol_name[5];

int	i_conv;
int start_symbol = 1295;
int	i_symbol = 1295;
const   int last_symbol = 46655;
int	tmp_symbol, min_tmp_symbol = 46655;

CNV_LIST	conv_list;

char	*calc_fam_number( int	l )
{
    int	i, j, k;

    i = l/fam_n_chars;
    j = i%fam_n_chars;
    k = l%fam_n_chars;
    i = i/fam_n_chars;

    if( i >= fam_n_chars )
        return NULL;

    fam_number[0] = fam_chars[i];
    fam_number[1] = fam_chars[j];
    fam_number[2] = fam_chars[k];

    return fam_number;
}

int	calc_number(	char	*a )
{
    int	i, j, k, num;

    i = ( a[3] < 65 ) ? a[3] - 48 : a[3] - 55;
    j = ( a[2] < 65 ) ? a[2] - 48 : a[2] - 55;
    k = ( a[1] < 65 ) ? a[1] - 48 : a[1] - 55;
    num = k*fam_n_chars*fam_n_chars + j*fam_n_chars + i;
    return num;
}

void	next_fam_symbol()
{
    i_symbol++;
    calc_fam_number( i_symbol );
    strcpy( &fam_symbol_name[1], fam_number );
}

void	prev_fam_symbol()
{
    tmp_symbol--;
    calc_fam_number( tmp_symbol );
    strcpy( &tmp_symbol_name[1], fam_number );
}



double	*poly_2_val(	double	*a,
			int     nu,
			int     nv,
			double  u,
			double  v,
			double  *coord )
{
    int	i_coord, lu, lv;
    double	c, d;
    double	*b;

    for( i_coord = 0; i_coord < 3; i_coord++ )
	{
            a = a + nu*nv;
            d = 0.;
            b = a;
            for( lv = nv; lv > 0; lv-- )
		{
                    c = *b--;
                    for( lu = nu-1; lu > 0; lu-- )
                        c = c*u + *b--;
                    d = d*v + c;
		}
            *coord++ = d;
	}
    return coord-3;
}

void	save_conv_name(	char	*vda_name,
			char	*fam_name )
{
    CNV_LIST	pair;

    pair = conv_list;
    while( pair->next != NULL )
        pair = pair->next;
    pair->next = (CNV_LIST)malloc( sizeof( struct cnv_list ) );
    pair = pair->next;
    i_conv++;
    pair->vda_name = malloc( 1 + strlen( vda_name ) );
    strcpy( pair->vda_name, vda_name );
    pair->fam_name = malloc( 1 + strlen( fam_name ) );
    strcpy( pair->fam_name, fam_name );
    pair->next = NULL;
}

void	vda_header2fam(	VDA_HEADER	header )
{
    printf( "\n** Header\n" );
    next_fam_symbol();
    printf( "TEXT !%4s\n", fam_symbol_name );
    while( ( header = header->next ) != NULL )
	{
            printf( "%-20s%s\n", header->key, header->value );
            if( !strcmp( header->key, "TOLERANZ           \0" ) ) {
                /*		  sscanf( header->value, "%lf%s", &gtol, &unit ); */
                sscanf( header->value, "%lf", &gtol );
            }
	}
    printf( "\n\n" );
}

void	vda_points2fam(	VDA_POINT	point )
{
    int	i;
    double*	coord;

    while( ( point = point->next ) != NULL )
	{
            printf( "** %-8s\n", point->name );
            next_fam_symbol();
            printf( "PNT  !%4s", fam_symbol_name );
            coord = point->c;
            for( i = 0; i < 3; i++, coord++ )
                printf( "%15.5f", *coord );
            printf( "\n\n" );

            save_conv_name( point->name, fam_symbol_name );
	}
}

void	vda_psets2fam(	VDA_PSET	pset )
{
    int	i, j;
    double* coord;
    char   tmp_point0_name[5], tmp_point1_name[5];

    while( ( pset = pset->next ) != NULL ) {
        printf( "** %-8s\n", pset->name );
        /*      next_fam_symbol(); */
        /*      printf( "SETO !%4s\n", fam_symbol_name ); */
        /*      tmp_symbol = last_symbol; */
        prev_fam_symbol();
        printf( "PNT  !%4s ", tmp_symbol_name );
        strcpy( tmp_point0_name, tmp_symbol_name );
        coord = pset->c;
        for( i = 0; i < 3; i++, coord++ )
            printf( "%15.5f", *coord );
        printf( "\n" );
        for( j = 1; j < pset->n; j++ ) {
            prev_fam_symbol();
            printf( "PNT  !%4s ", tmp_symbol_name );
            strcpy( tmp_point1_name, tmp_symbol_name );
            for( i = 0; i < 3; i++, coord++ )
                printf( "%15.5f", *coord );
            printf( "\n" );
            /*	prev_fam_symbol(); */
            next_fam_symbol();
            printf("LINE !%4s %c%4s %c%4s\n", fam_symbol_name,
                   '%', tmp_point0_name, '%', tmp_point1_name );
            strcpy( tmp_point0_name, tmp_point1_name );
        }
        /*      printf( "SETC %c%4s\n", '%', fam_symbol_name ); */
        printf( "\n" );
        save_conv_name( pset->name, fam_symbol_name );
        min_tmp_symbol = IMIN( tmp_symbol, min_tmp_symbol );
    }
}

void	vda_mdis2fam(	VDA_MDI	mdi )
{
    int	i, j;
    double*	coord;

    while( ( mdi = mdi->next ) != NULL )
	{
            printf( "** %-8s\n", mdi->name );
            coord = mdi->c;
            for( j = 0; j < mdi->n; j++ )
		{
                    printf( "PNT  ! " );
                    for( i = 0; i < 6; i++, coord++ )
                        printf( "%15.5f", *coord );
                    printf( "\n\n" );
		}
	}
}


void	vda_circles2fam(	VDA_CIRCLE	circle )
{
    int	i, j, k;
    double	*p, r, alpha, beta, sa, sb, a0[3], b0[3],
        a[3], b[3], c[3], rot[3][3];
    double	ab;
    char	center_point_name[5], start_point_name[5], end_point_name[5];
    char    tmp_point_name[5];
    char	combi_line[40] = "LCMB !                                 \0";

    while( ( circle = circle->next ) != NULL )
	{
            printf( "** %-8s\n", circle->name );
            /* 		i_point = -1; */
            tmp_symbol = last_symbol;

            p = circle->params;
            for( i = 0; i < 3; i++ )
                c[i] = *p++;
            r = *p++;
            for( j = 0; j < 2; j++ )
                for( i = 0; i < 3; i++ )
                    rot[i][j] = *p++;
            alpha = M_PI*(*p++)/180.0;
            beta  = M_PI*(*p)/180.0;

            rot[0][2] = rot[1][0]*rot[2][1] - rot[2][0]*rot[1][1];
            rot[1][2] = rot[2][0]*rot[0][1] - rot[0][0]*rot[2][1];
            rot[2][2] = rot[0][0]*rot[1][1] - rot[1][0]*rot[0][1];

            a[0] = r*cos( alpha );
            a[1] = r*sin( alpha );
            b[0] = r*cos( beta );
            b[1] = r*sin( beta );
            a[2] = b[2] = 0.0;

            for( i = 0; i < 3; i++ )
		{
                    sa = sb = 0.0;
                    for( j = 0; j < 3; j++ )
			{
                            sa = sa + rot[i][j]*a[j];
                            sb = sb + rot[i][j]*b[j];
			}
                    a0[i] = c[i] + sa;
                    b0[i] = c[i] + sb;
		}
            a[0] = b0[0] - a0[0];
            a[1] = b0[1] - a0[1];
            a[2] = b0[2] - a0[2];
            ab = sqrt( a[0]*a[0] + a[1]*a[1] + a[2]*a[2] );

            prev_fam_symbol();
            strcpy( center_point_name, tmp_symbol_name );
            printf( "PNT  !%4s%15.5f%15.5f%15.5f\n",
                    tmp_symbol_name, c[0]*scale, c[1]*scale, c[2]*scale );
            prev_fam_symbol();
            strcpy( start_point_name, tmp_symbol_name );
            printf( "PNT  !%4s%15.5f%15.5f%15.5f\n",
                    tmp_symbol_name,
                    a0[0]*scale, a0[1]*scale, a0[2]*scale );
            if( beta-alpha < M_PI )
		{
                    prev_fam_symbol();
                    strcpy( end_point_name, tmp_symbol_name );
                    printf( "PNT  !%4s%15.5f%15.5f%15.5f\n",
                            tmp_symbol_name,
                            b0[0]*scale, b0[1]*scale, b0[2]*scale );
                    next_fam_symbol();
                    printf( "LINE !%4s %c%4s %c%4s %c%4s\n",
                            fam_symbol_name,
                            '%', start_point_name, '%', end_point_name,
                            '%', center_point_name );
                    printf( "\n" );
                    save_conv_name( circle->name, fam_symbol_name );
		}
            else if( beta-alpha < 2.0*M_PI )
		{
                    for( i = 6; i < 39; i++ )
                        combi_line[i] = ' ';
                    prev_fam_symbol();
                    strcpy( end_point_name, tmp_symbol_name );
                    printf( "PNT  !%4s%15.5f%15.5f%15.5f\n",
                            tmp_symbol_name,
                            b0[0]*scale, b0[1]*scale, b0[2]*scale );

                    alpha = ( beta + alpha )/2.;
                    a[0] = r*cos( alpha );
                    a[1] = r*sin( alpha );
                    a[2] = 0.0;
                    for( i = 0; i < 3; i++ )
			{
                            sa = 0.0;
                            for( j = 0; j < 3; j++ )
                                sa = sa + rot[i][j]*a[j];
                            a0[i] = c[i] + sa;
			}
                    prev_fam_symbol();
                    strcpy( tmp_point_name, tmp_symbol_name );
                    printf( "PNT  !%4s%15.5f%15.5f%15.5f\n",
                            tmp_point_name,
                            a0[0]*scale, a0[1]*scale, a0[2]*scale );
                    prev_fam_symbol();
                    printf( "LINE !%4s %c%4s %c%4s %c%4s\n",
                            tmp_symbol_name,
                            '%', start_point_name, '%', tmp_point_name,
                            '%', center_point_name );
                    combi_line[11] = '%';
                    for( i = 12, j = 0; j < 4; i++, j++ )
                        combi_line[i] = tmp_symbol_name[j];
                    prev_fam_symbol();
                    printf( "LINE !%4s %c%4s %c%4s %c%4s\n",
                            tmp_symbol_name,
                            '%', tmp_point_name, '%', end_point_name,
                            '%', center_point_name );
                    combi_line[17] = '%';
                    for( i = 18, j = 0; j < 4; i++, j++ )
                        combi_line[i] = tmp_symbol_name[j];
                    next_fam_symbol();
                    for( i = 6, j = 0; j < 4; i++, j++ )
                        combi_line[i] = fam_symbol_name[j];
                    printf( "%-s\n", combi_line );
                    printf( "\n" );
                    save_conv_name( circle->name, fam_symbol_name );
		}
            else
		{
                    strcpy( end_point_name, start_point_name );
                    beta = alpha;
                    for( k = 0; k < 3; k++ )
			{
                            beta += M_PI_2;
                            b[0] = r*cos( beta );
                            b[1] = r*sin( beta );
                            b[2] = 0.0;
                            for( i = 0; i < 3; i++ )
				{
                                    sa = sb = 0.0;
                                    for( j = 0; j < 3; j++ )
                                        sb = sb + rot[i][j]*b[j];
                                    b0[i] = c[i] + sb;
				}
                            prev_fam_symbol();
                            strcpy( tmp_point_name, tmp_symbol_name );
                            printf( "PNT  !%4s%15.5f%15.5f%15.5f\n",
                                    tmp_point_name,
                                    b0[0]*scale, b0[1]*scale, b0[2]*scale );
                            prev_fam_symbol();
                            printf( "LINE !%4s %c%4s %c%4s %c%4s\n",
                                    tmp_symbol_name,
                                    '%', start_point_name, '%', tmp_point_name,
                                    '%', center_point_name );
                            strcpy( start_point_name, tmp_point_name );
                            combi_line[11+6*k] = '%';
                            for( i = 12+6*k, j = 0; j < 4; i++, j++ )
                                combi_line[i] = tmp_symbol_name[j];
			}
                    prev_fam_symbol();
                    printf( "LINE !%4s %c%4s %c%4s %c%4s\n",
                            tmp_symbol_name, '%', tmp_point_name,
                            '%', end_point_name, '%', center_point_name );
                    combi_line[29] = '%';
                    for( i = 30, j = 0; j < 4; i++, j++ )
                        combi_line[i] = tmp_symbol_name[j];
                    next_fam_symbol();
                    for( i = 6, j = 0; j < 4; i++, j++ )
                        combi_line[i] = fam_symbol_name[j];
                    printf( "%-s\n", combi_line );
                    printf( "\n" );
                    save_conv_name( circle->name, fam_symbol_name );
		}
            min_tmp_symbol = IMIN( tmp_symbol, min_tmp_symbol );
	}
}

void	vda_strips2fam( VDA_CURVE curve )
{
    int	i, k, ns, u;
    int  	u_order;
    char	first_point_name[5], last_point_name[5], tmp_nurl_name[5];
    double	*a, *x, *y, *z;
    double	**LT_i, *c;
    POLY_1	poly;

    LT_i = double_matrix( 0, max_u_order-1, 0, max_u_order-1 );
    c    = double_vector( 0, max_u_order-1 );
    x    = double_vector( 0, max_u_order-1 );
    y    = double_vector( 0, max_u_order-1 );
    z    = double_vector( 0, max_u_order-1 );

    while( curve->next != NULL ) {
	curve = curve->next;
	printf( "** %-8s\n", curve->name );
	next_fam_symbol();
	printf( "SETO !%4s\n", fam_symbol_name );

	ns = curve->ns;
	poly = curve->poly;
	u_order = 0;
	while( ( poly = poly->next ) != NULL )
	    u_order = LMAX( u_order, (long)poly->nu );
	poly = curve->poly;
	for( k = 0; k < ns ; k++ ) {
	    for( i = 0; i < u_order; i++ )
		x[i] = y[i] = z[i] = 0.0;
	    poly = poly->next;
	    u = poly->nu;
	    a = poly->a;
	    for( i = 0; i < u; i++ )
		x[i] = *(a++);
	    for( i = 0; i < u; i++ )
		y[i] = *(a++);
	    for( i = 0; i < u; i++ )
		z[i] = *(a++);

	    pol2bez_curve( x, u_order, LT_i, c );
	    pol2bez_curve( y, u_order, LT_i, c );
	    pol2bez_curve( z, u_order, LT_i, c );

	    tmp_symbol = last_symbol;
	    prev_fam_symbol();
	    strcpy( first_point_name, tmp_symbol_name );
	    tmp_symbol = last_symbol;
	    for( i = 0; i < u_order; i++ ) {
		prev_fam_symbol();
		printf( "PNT  !%4s%15.5f%15.5f%15.5f\n",
			tmp_symbol_name,
			scale*x[i], scale*y[i], scale*z[i] );
	    }
	    strcpy( last_point_name, tmp_symbol_name );
	    prev_fam_symbol();
	    strcpy( tmp_nurl_name, tmp_symbol_name );
	    printf( "NURL !%4s DEFINE  %c%4s %c%4s %4d %4d %4s\n",
		    tmp_nurl_name, '%', first_point_name, '%',
		    last_point_name, u_order-1, u_order, "PBEZ" );
	    tmp_symbol = last_symbol;
	    for( i = 0; i < u_order; i++ ) {
		prev_fam_symbol();
		printf( "NURL %c%4s CONTROL  %4d %c%4s\n",
			'%', tmp_nurl_name, i, '%', tmp_symbol_name );
	    }
	    printf( "NURL %c%4s END\n", '%', tmp_nurl_name );
	}
	printf( "SETC %c%4s\n", '%', fam_symbol_name );
	printf( "MERG P %c%4s\n\n", '%', fam_symbol_name );
	save_conv_name( curve->name, fam_symbol_name );
	min_tmp_symbol = IMIN( tmp_symbol, min_tmp_symbol );
    }
    free_double_vector( x,    0, max_u_order-1 );
    free_double_vector( y,    0, max_u_order-1 );
    free_double_vector( z,    0, max_u_order-1 );
    free_double_vector( c,    0, max_u_order-1 );
    free_double_matrix( LT_i, 0, max_u_order-1, 0, max_u_order-1 );
}

void	vda_curves2fam(	VDA_CURVE	curve )
{
    int	i, j, k, ns, u;
    int 	u_order;
    char	first_point_name[5];
    double	start[3], end[3];
    double	*a, *x, *y, *z;
    double	**LT_i, *c;
    POLY_1	poly;

    LT_i = double_matrix( 0, max_u_order-1, 0, max_u_order-1 );
    c    = double_vector( 0, max_u_order-1 );
    x    = double_vector( 0, max_u_order-1 );
    y    = double_vector( 0, max_u_order-1 );
    z    = double_vector( 0, max_u_order-1 );

    while( curve->next != NULL ) {
	tmp_symbol = last_symbol;
	/* 		i_point = -1; */
	curve = curve->next;
	printf( "** %-8s\n", curve->name );

	ns = curve->ns;
	poly = curve->poly;
	u_order = 0;
	while( ( poly = poly->next ) != NULL )
	    u_order = LMAX( u_order, (long)poly->nu );
	poly = curve->poly;

	/* 		tmp_point = i_point; */
	prev_fam_symbol();
	strcpy( first_point_name, tmp_symbol_name );

	for( k = 0; k < ns ; k++ ) {
	    for( i = 0; i < u_order; i++ )
		x[i] = y[i] = z[i] = 0.0;
	    poly = poly->next;
	    u = poly->nu;
	    a = poly->a;
	    for( i = 0; i < u; i++ )
		x[i] = *(a++);
	    for( i = 0; i < u; i++ )
		y[i] = *(a++);
	    for( i = 0; i < u; i++ )
		z[i] = *(a++);

	    pol2bez_curve( x, u_order, LT_i, c );
	    pol2bez_curve( y, u_order, LT_i, c );
	    pol2bez_curve( z, u_order, LT_i, c );
	    for( i = 0; i < u_order-1; i++ ) {
		printf( "PNT  !%4s%15.5f%15.5f%15.5f\n",
			tmp_symbol_name,
			scale*x[i], scale*y[i], scale*z[i] );
		prev_fam_symbol();
	    }
	    if( k == 0 ) {
		start[0] = x[0];
                start[1] = y[0];
                start[2] = z[0];
	    }
	}
	end[0] = x[u_order-1];
	end[1] = y[u_order-1];
	end[2] = z[u_order-1];
	printf( "PNT  !%4s%15.5f%15.5f%15.5f\n",
		tmp_symbol_name,
		scale*end[0], scale*end[1], scale*end[2] );

	/* i_point = tmp_point; */
	/* 		max_point = ( i_point > max_point ) ? i_point : max_point; */
	next_fam_symbol();
	printf( "NURL !%4s DEFINE  %c%4s %c%4s %4d %4d %4d\n",
		fam_symbol_name, '%', first_point_name, '%',
		tmp_symbol_name,
		u_order-1, ns*(u_order-1)+1, (ns+1)*(u_order-1)+2 );

	/* 		i_point = -1; */
	tmp_symbol = last_symbol;
	for( k = 0; k < ns ; k++ ) {
	    for( i = 0; i < u_order-1; i++ ) {
		prev_fam_symbol();
		printf( "NURL %c%4s CONTROL  %4d %c%4s\n",
			'%', fam_symbol_name,
			k*(u_order-1)+i+1, '%', tmp_symbol_name );
	    }
	}
	prev_fam_symbol();
	printf( "NURL %c%4s CONTROL  %4d %c%4s\n",
		'%', fam_symbol_name, ns*(u_order-1)+1,
		'%', tmp_symbol_name );

	printf( "NURL %c%4s KNOT     %4d  %4d\n",
		'%', fam_symbol_name, 1,0);
	for( j = 0; j <= ns; j++ )
	    for( i = 1; i < u_order; i++ )
		printf( "NURL %c%4s KNOT     %4d  %4d\n",
			'%', fam_symbol_name,
			(j*(u_order-1))+i+1,j);
	printf( "NURL %c%4s KNOT     %4d  %4d\n",
		'%', fam_symbol_name, (ns+1)*(u_order-1)+2, ns );
	printf( "NURL %c%4s END\n\n", '%', fam_symbol_name );

	save_conv_name( curve->name, fam_symbol_name );

	min_tmp_symbol = IMIN( tmp_symbol, min_tmp_symbol );
    }
    free_double_vector( x,    0, max_u_order-1 );
    free_double_vector( y,    0, max_u_order-1 );
    free_double_vector( z,    0, max_u_order-1 );
    free_double_vector( c,    0, max_u_order-1 );
    free_double_matrix( LT_i, 0, max_u_order-1, 0, max_u_order-1 );
}

void	print_control_points(	char *nurs_symbol,
				int     u_min,
				int     u_max,
				int     v_min,
				int     v_max,
				int     k,
				int     l,
				int     u_order,
				int     v_order,
				double  **X,
				double  **Y,
				double  **Z )
{
    int	i, j;
    int tmp_point;

    tmp_point = tmp_symbol;
    for( j = v_min; j <= v_max; j++ ) {
        for( i = u_min; i <= u_max; i++ ) {
            prev_fam_symbol();
            printf( "PNT  !%4s%15.5f%15.5f%15.5f\n",
                    tmp_symbol_name,
                    scale*X[i][j], scale*Y[i][j], scale*Z[i][j] );
        }
    }
    tmp_symbol = tmp_point;
    for( j = v_min; j <= v_max; j++ ) {
        for( i = u_min; i <= u_max; i++ ) {
            prev_fam_symbol();
            printf( "NURS %c%4s CONTROL %4d %4d %c%4s\n",
                    '%', nurs_symbol,
                    k*(u_order-1)+i+1, l*(v_order-1)+j+1,
                    '%', tmp_symbol_name );
        }
    }
}

void	vda_patches2fam( VDA_SURFACE surface )
{
    int	i, j, k, l, ns, nt, u, v;
    int 	u_order, v_order;
    double	*a;
    double  **X, **Y, **Z;
    double **MT_i, **N_i, **b;
    POLY_2	poly;
    char tmp_nurs_name[5];
    /* 	COORD_3D	coord; */

    MT_i = double_matrix( 0, max_u_order-1, 0, max_u_order-1 );
    N_i  = double_matrix( 0, max_v_order-1, 0, max_v_order-1 );
    b    = double_matrix( 0, max_u_order-1, 0, max_v_order-1 );
    X    = double_matrix( 0, max_u_order-1, 0, max_v_order-1 );
    Y    = double_matrix( 0, max_u_order-1, 0, max_v_order-1 );
    Z    = double_matrix( 0, max_u_order-1, 0, max_v_order-1 );

    while( ( surface = surface->next ) != NULL ) {
	/* 		i_point = -1; */
	ns = surface->ns;
	nt = surface->nt;
	poly = surface->poly;
	u_order = v_order = 0;
	while( ( poly = poly->next ) != NULL ) {
	    u_order = LMAX( u_order, (long)poly->nu );
	    v_order = LMAX( v_order, (long)poly->nv );
	}
	poly = surface->poly;

	printf( "** %-8s\n", surface->name );
	next_fam_symbol();
	printf( "SETO !%4s\n", fam_symbol_name );

	for( l = 0; l < nt; l++ ) {
	    for( k = 0; k < ns; k++ ) {
		poly = poly->next;
		u = poly->nu;
		v = poly->nv;

		for( j = 0; j < v_order; j++ )
		    for( i = 0; i < u_order; i++ )
			X[i][j] = 0.0;
		for( j = 0; j < v_order; j++ )
		    for( i = 0; i < u_order; i++ )
			Y[i][j] = 0.0;
		for( j = 0; j < v_order; j++ )
		    for( i = 0; i < u_order; i++ )
			Z[i][j] = 0.0;

		a = poly->a;
		for( j = 0; j < v; j++ )
		    for( i = 0; i < u; i++ )
			X[i][j] = *(a++);
		for( j = 0; j < v; j++ )
		    for( i = 0; i < u; i++ )
			Y[i][j] = *(a++);
		for( j = 0; j < v; j++ )
		    for( i = 0; i < u; i++ )
			Z[i][j] = *(a++);

		pol2bez_surf( X, u_order, v_order, MT_i, N_i, b );
		pol2bez_surf( Y, u_order, v_order, MT_i, N_i, b );
		pol2bez_surf( Z, u_order, v_order, MT_i, N_i, b );

		tmp_symbol = last_symbol;
		prev_fam_symbol();
		strcpy( tmp_nurs_name, tmp_symbol_name );
		printf( "NURS !%4s DEFINE  %4d %4d %4d %4d %4s %4s\n",
			tmp_nurs_name, u_order-1, v_order-1,
			u_order, v_order, "PBEZ", "PBEZ" );
		print_control_points( tmp_nurs_name,
				      0, u_order-1, 0, v_order-1, 0, 0,
				      u_order, v_order, X, Y, Z );
		printf( "NURS %c%4s END\n", '%', tmp_nurs_name );
		min_tmp_symbol = IMIN( tmp_symbol, min_tmp_symbol );
	    }
	}
	printf( "SETC %c%4s\n", '%', fam_symbol_name );
	printf( "MERG P %c%4s\n", '%', fam_symbol_name );
	printf( "MERG L %c%4s\n\n", '%', fam_symbol_name );
	save_conv_name( surface->name, fam_symbol_name );
    }
    free_double_matrix( X,    0, max_u_order-1, 0, max_v_order-1 );
    free_double_matrix( Y,    0, max_u_order-1, 0, max_v_order-1 );
    free_double_matrix( Z,    0, max_u_order-1, 0, max_v_order-1 );
    free_double_matrix( MT_i, 0, max_u_order-1, 0, max_u_order-1 );
    free_double_matrix( N_i,  0, max_v_order-1, 0, max_v_order-1 );
    free_double_matrix( b,    0, max_u_order-1, 0, max_v_order-1 );
}

void	vda_surfaces2fam(	VDA_SURFACE	surface )
{
    int	i, j, k, l, ns, nt, u, v;
    int 	u_order, v_order;
    double	*a;
    double  **X, **Y, **Z;
    double **MT_i, **N_i, **b;
    POLY_2	poly;
    /* 	COORD_3D	coord; */

    MT_i = double_matrix( 0, max_u_order-1, 0, max_u_order-1 );
    N_i  = double_matrix( 0, max_v_order-1, 0, max_v_order-1 );
    b    = double_matrix( 0, max_u_order-1, 0, max_v_order-1 );
    X    = double_matrix( 0, max_u_order-1, 0, max_v_order-1 );
    Y    = double_matrix( 0, max_u_order-1, 0, max_v_order-1 );
    Z    = double_matrix( 0, max_u_order-1, 0, max_v_order-1 );

    while( ( surface = surface->next ) != NULL ) {
      tmp_symbol = last_symbol;
      /* 		i_point = -1; */
      ns = surface->ns;
      nt = surface->nt;
      poly = surface->poly;
      u_order = v_order = 0;
      while( ( poly = poly->next ) != NULL ) {
        u_order = LMAX( u_order, (long)poly->nu );
        v_order = LMAX( v_order, (long)poly->nv );
      }
      poly = surface->poly;
      
      printf( "** %-8s\n", surface->name );
      next_fam_symbol();
      printf( "NURS !%4s DEFINE  %4d %4d %4d %4d %4d %4d\n",
              fam_symbol_name, u_order-1, v_order-1,
              ns*(u_order-1)+1, nt*(v_order-1)+1,
              (ns+1)*(u_order-1)+2, (nt+1)*(v_order-1)+2 );
      
      for( l = 0; l < nt; l++ ) {
        for( k = 0; k < ns; k++ ) {
          poly = poly->next;
          u = poly->nu;
          v = poly->nv;

          for( j = 0; j < v_order; j++ )
            for( i = 0; i < u_order; i++ )
              X[i][j] = 0.0;
          for( j = 0; j < v_order; j++ )
            for( i = 0; i < u_order; i++ )
              Y[i][j] = 0.0;
          for( j = 0; j < v_order; j++ )
            for( i = 0; i < u_order; i++ )
              Z[i][j] = 0.0;

          a = poly->a;
          for( j = 0; j < v; j++ )
            for( i = 0; i < u; i++ )
              X[i][j] = *(a++);
          for( j = 0; j < v; j++ )
            for( i = 0; i < u; i++ )
              Y[i][j] = *(a++);
          for( j = 0; j < v; j++ )
            for( i = 0; i < u; i++ )
              Z[i][j] = *(a++);
          
          pol2bez_surf( X, u_order, v_order, MT_i, N_i, b );
          pol2bez_surf( Y, u_order, v_order, MT_i, N_i, b );
          pol2bez_surf( Z, u_order, v_order, MT_i, N_i, b );

          print_control_points( fam_symbol_name,
                                0, u_order-2, 0, v_order-2, k, l,
                                u_order, v_order, X, Y, Z );
          if( l == nt-1 )
            print_control_points( fam_symbol_name,
                                  0, u_order-2, v_order-1,
                                  v_order-1, k, nt-1,
                                  u_order, v_order, X, Y, Z );
        }
        print_control_points( fam_symbol_name,
                              u_order-1, u_order-1, 0, v_order-2,
                              ns-1, l, u_order, v_order, X, Y, Z );
      }
      print_control_points( fam_symbol_name,
                            u_order-1, u_order-1, v_order-1, v_order-1,
                            ns-1, nt-1, u_order, v_order, X, Y, Z );
      
      printf( "NURS %c%4s KNOT  U %4d %4d\n", '%',
              fam_symbol_name, 1, 0 );
      for( j = 0; j <= ns; j++ )
        for( i = 1; i < u_order; i++ )
          printf( "NURS %c%4s KNOT  U %4d %4d\n",
                  '%', fam_symbol_name, (j*(u_order-1))+i+1, j );
      printf( "NURS %c%4s KNOT  U %4d %4d\n", '%', fam_symbol_name,
              (ns+1)*(u_order-1)+2, ns );
      
      printf( "NURS %c%4s KNOT  V %4d %4d\n", '%',
              fam_symbol_name, 1, 0 );
      for( j = 0; j <= nt; j++ )
        for( i = 1; i < v_order; i++ )
          printf( "NURS %c%4s KNOT  V %4d %4d\n",
                  '%', fam_symbol_name, (j*(v_order-1))+i+1, j );
      printf( "NURS %c%4s KNOT  V %4d %4d\n", '%', fam_symbol_name,
              (nt+1)*(v_order-1)+2, nt );
      
      printf( "NURS %c%4s END\n\n", '%', fam_symbol_name );

      save_conv_name( surface->name, fam_symbol_name );
      
      min_tmp_symbol = IMIN( tmp_symbol, min_tmp_symbol );
    }
    free_double_matrix( X,    0, max_u_order-1, 0, max_v_order-1 );
    free_double_matrix( Y,    0, max_u_order-1, 0, max_v_order-1 );
    free_double_matrix( Z,    0, max_u_order-1, 0, max_v_order-1 );
    free_double_matrix( MT_i, 0, max_u_order-1, 0, max_u_order-1 );
    free_double_matrix( N_i,  0, max_v_order-1, 0, max_v_order-1 );
    free_double_matrix( b,    0, max_u_order-1, 0, max_v_order-1 );
}

double	*cons_point(	VDA_CONS	vcons,
			VDA_SURFACE	vsurf,
			double		param,
			double		*pnt )
{
    int		j, k, lu, nu, nv;
    double		*a, *b, c, s, t;
    CONS_CURVE	vcurve;
    POLY_2		vpoly;

    a = vcons->segment_params + 1;
    vcurve = vcons->cons_curve->next;
    k = 1;
    while( param > *a && k < vcons->n_segments ) {
      a++; k++;
      vcurve = vcurve->next;
    }
    nu = vcurve->order;
    t = ( param - *(a-1) )/( *a - *(a-1) );
    a = vcurve->coeff + nu - 1;
    b = a;
    c = *b--;
    for( lu = nu-1; lu > 0; lu-- )
        c = c*t + *b--;
    s = c;

    a = vcurve->coeff + 2*nu - 1;
    b = a;
    c = *b--;
    for( lu = nu-1; lu > 0; lu-- )
        c = c*t + *b--;
    t = c;

    a = vsurf->a + 1;
    vpoly = vsurf->poly->next;
    k = 1;
    while( s > *a && k < vsurf->ns ) {
      a++; k++;
      vpoly = vpoly->next;
    };
    s = ( s - *(a-1) )/( *a - *(a-1) );
    a = vsurf->a + vsurf->ns + 2;
    k = 1;
    while( t > *a && k < vsurf->nt ) {
      a++; k++;
      for( j = 0; j < vsurf->ns; j++ )
        vpoly = vpoly->next;
    };
    t = ( t - *(a-1) )/( *a - *(a-1) );
    nu = vpoly->nu;
    nv = vpoly->nv;
    pnt = poly_2_val( vpoly->a-1, nu, nv, s, t, &pnt[0] );

    return pnt;
}

void	vda_face2fam(	VDA_FACE	face )
{
    int		i;
    FACE_LOOP	loop;
    FACE_CONS	face_cons;
    CNV_LIST	conv;
    char		outer_loop[5];

    //printf( "SETR %c%s ALL\n\n", '%', tmp_set );
    while( ( face = face->next ) != NULL ) {
            tmp_symbol = last_symbol;
            printf( "** %-8s\n", face->name );

            next_fam_symbol();
            conv = conv_list;
            while( conv->next != NULL ) {
                conv = conv->next;
                if( !strcmp( face->surf->name, conv->vda_name ) )
                    break;
            }	
            printf( "GSUR !%4s + %%%4s ", fam_symbol_name, conv->fam_name );

            loop = face->loop->next;
            face_cons = loop->face_cons;
            i = 0;
            while( ( face_cons = face_cons->next ) != NULL ) {
                    conv = conv_list;
                    while( conv->next != NULL ) {
                        conv = conv->next;
                        if( !strcmp( face_cons->cons->curve->name,
                                     conv->vda_name ) )
                            break;
                    }
                    printf( "+ %c%4s ", '%', conv->fam_name );
                    if( (i++)%7 == 6 && face_cons->next != NULL )
                        printf("\nGSUR %c%4s ADD     ", '%', fam_symbol_name);
            }
            printf( "\n" );
            strcpy( outer_loop, fam_symbol_name );
            save_conv_name( face->name, fam_symbol_name );

            while( ( loop = loop->next ) != NULL ) {
              /* next_fam_surface(); */
              conv = conv_list;
              while( conv->next != NULL ) {
                conv = conv->next;
                if( !strcmp( face->surf->name, conv->vda_name ) )
                  break;
              }
              printf( "** inner loop\n" );
              printf( "GSUR %%%4s ADD     ", fam_symbol_name );
              
              face_cons = loop->face_cons;
              i = 0;
              while( ( face_cons = face_cons->next ) != NULL ) {
                conv = conv_list;
                while( conv->next != NULL ) {
                  conv = conv->next;
                  if(!strcmp(face_cons->cons->curve->name,
                             conv->vda_name ) )
                    break;
                }	
                printf( "+ %c%4s ", '%', conv->fam_name );
                if( (i++)%7 == 6 && face_cons->next != NULL )
                  printf( "\nGSUR %c%4s ADD     ", '%', fam_symbol_name );
              }
              printf( "\n" );
            }

            //printf( "SETA %c%s s %c%4s\n", '%', tmp_set, '%', fam_symbol_name );
            //printf( "COMP %c%s D\n", '%', tmp_set );
            //printf( "MERG p %c%s\n", '%', tmp_set );
            //printf( "ORIENT %c%s\n", '%', tmp_set );
            //printf( "DEL se %c%s\n", '%', tmp_set );
            /*
            if( strlen( face->identifier ) ) {
                prev_fam_symbol();
                printf( "TEXT !%s\n%s\n\n", tmp_symbol_name, face->identifier );
                printf( "ASG  %c%s %c%s\n",
                        '%', fam_symbol_name, '%', tmp_symbol_name );
            }
            printf( "\n" );
            */
            min_tmp_symbol = IMIN( tmp_symbol, min_tmp_symbol );
	}
}

void	vda_groups2fam(	VDA_GROUP	group )
{
    int		i, j;
    CNV_LIST	conv;
    VDA_GROUP tmp_group;
    int set_symbol = i_symbol;

    tmp_group = group;
    /*
      fprintf( stderr, "*famgui.modeltools:" );
      while( ( tmp_group = tmp_group->next ) != NULL ) {
      fprintf( stderr, " modeltool_%s", tmp_group->name );
      }
      fprintf( stderr, "\n" );
      tmp_group = group;
      while( ( tmp_group = tmp_group->next ) != NULL ) {
      next_fam_symbol();
      fprintf( stderr, "*famgui.modeltool_%s: {} {PLOT SC %c%s} {}\n",
      tmp_group->name, '%', fam_symbol_name );
      }
      fprintf( stderr, "*famgui.modeltoolboxes: modeltoolbox_Model_Tools\n" );
      fprintf( stderr, "*famgui.modeltoolbox_Model_Tools:" );
      tmp_group = group;
      while( ( tmp_group = tmp_group->next ) != NULL ) {
      fprintf( stderr, " modeltool_%s", tmp_group->name );
      }
      fprintf( stderr, "\n" );
    */

    i_symbol = set_symbol;
    while( ( group = group->next ) != NULL ) {
            printf( "** %-8s\n", group->name );
            next_fam_symbol();
            for( j = 0; j < group->n_objects/8; j++ ) {
                if( j == 0 )
                    printf( "SETA !%4s", fam_symbol_name );
                else
                    printf( "SETA %%%4s", fam_symbol_name );
                for( i = 0; i < 8; i++ ) {
                    conv = conv_list;
                    while( conv->next != NULL ) {
                        conv = conv->next;
                        if( !strcmp( group->object_list[j*8+i],
                                     conv->vda_name ) )
                            printf(" %%%s",conv->fam_name);
                    }
                }	
                printf( "\n" );
            }
            if( group->n_objects%8 ) {
                if( j == 0 )
                    printf( "SETA !%4s", fam_symbol_name );
                else
                    printf( "SETA %%%4s", fam_symbol_name );
                for( i = 0; i < group->n_objects%8; i++ ) {
                    conv = conv_list;
                    while( conv->next != NULL ) {
                        conv = conv->next;
                        if( !strcmp( group->object_list[j*8+i],
                                     conv->vda_name ) )
                            printf(" %%%s",conv->fam_name);
                    }
                }	
                printf( "\n" );
            }
            printf( "SETA %c%s %c%4s\n", '%', tmp_set, '%', fam_symbol_name );
            printf( "COMP %c%s D\n", '%', tmp_set );
            printf( "MERG P %c%s\n", '%', tmp_set );
            printf( "MERG L %c%s\n", '%', tmp_set );
            printf( "SETR %c%s ALL\n", '%', tmp_set );
            printf( "COMP %c%s D\n", '%', fam_symbol_name );
            printf( "\n" );
	}
    i_symbol = set_symbol;
}

void	delete_symbols()
{
    int	i;
    char c = fam_symbol_name[0];

    for( i = start_symbol+1; i <= i_symbol; i++ )
	printf( "DEL SB %c%3s\n", c, calc_fam_number(i) );
    for( i = min_tmp_symbol; i < last_symbol; i++ )
	printf( "DEL SB %c%3s\n", c, calc_fam_number(i) );
}

void vda2fam( VDA model, VDA_SHOW show )
{
    int	first_symbol;
    VDA_SET	set = model->set->next;
    char	set_name[5], old_model[5];

    strcpy( fam_number, "   \0" );
    strcpy( fam_symbol_name, show->symbol );
    strcpy( tmp_symbol_name, show->symbol );

    first_symbol = i_symbol;
    i_conv    = -1;
    tmp_symbol = last_symbol + 1;
    prev_fam_symbol();
    strcpy( tmp_set, tmp_symbol_name );
    next_fam_symbol();
    strcpy( set_name, fam_symbol_name );
    next_fam_symbol();
    strcpy( old_model, fam_symbol_name );
    conv_list = (CNV_LIST)malloc( sizeof( struct cnv_list ) );
    conv_list->next = NULL;

    printf( "PROC ECHO OFF\n" );

    printf( "SETA !%s CALLED ALL I IZZZ\n", tmp_set );
    printf( "RNAM ALL %c%s\n", '%', tmp_set );
    printf( "SETR %c%s ALL\n\n", '%', tmp_set );

    printf( "SETA !%s ALL\n", old_model );
    printf( "** PROC GLOK %c%s\n\n", '%', old_model );

    printf( "CSEL GLOB\n" );

    printf( "SETO !%s\n", set_name );
    if( show->header )
        vda_header2fam( set->header );
    printf( "PROC GTOL %8.5f\n\n", gtol );
    if( show->points )
        vda_points2fam( set->point );
    if( show->points || show->curves )
	vda_psets2fam( set->pset );
    /*	if( show->mdis )
        vda_mdis2fam( set->mdi ); */
    if( show->curves )
        vda_circles2fam( set->circle );
    if( show->patches )
        vda_strips2fam( set->curve );
    if( show->curves )
        vda_curves2fam( set->curve );
    if( show->patches )
        vda_patches2fam( set->surface );
    if( show->surfaces )
        vda_surfaces2fam( set->surface );
    if( show->faces )
        vda_face2fam( set->face );
    if( show->layers )
        vda_groups2fam( set->group );
    printf( "SETC %c%s\n\n", '%', set_name );

    printf( "PROC ECHO ON\n" );
}
